from lyriks.lyrics import Lyrics
from lyriks.mb_client import Release
from . import Provider
from .api import qqm_api
from .api.qqm_api import QQMSong
from .registry import register_provider


@register_provider('qq', 'qqm', 'qqmusic')
class QQMusic(Provider):
    """
    Provider for QQ Music.
    """

    def fetch_recording_lyrics(self, track_release: Release, recording_mbid: str) -> Lyrics | None:
        # Resolve album
        qqm_songs = self.get_mapped_provider_songs(
            track_release,
            lambda r: r.extract_url_str(r'https://y.qq.com/n/ryqq(?:_v2)?/albumDetail/(\w+)'),
            qqm_api.get_album_songs,
        )
        if not qqm_songs:
            return None

        # Get song for recording
        qqm_song = qqm_songs.get(recording_mbid)
        if not qqm_song:
            return None

        # Fetch lyrics
        return self.fetch_provider_song_lyrics(qqm_song)

    def fetch_song_by_id(self, song_id: int) -> QQMSong | None:
        return qqm_api.get_song_info(song_id)

    def fetch_provider_song_lyrics(self, song: QQMSong) -> Lyrics | None:
        return qqm_api.get_song_lyrics(song)

    @property
    def provider_domain(self) -> str:
        return 'y.qq.com'



import os
from dataclasses import dataclass

from lyriks.util import format_lrc_timestamp


def opener(path: str, flags: int) -> int:
    return os.open(path, flags, 0o644)


@dataclass
class Lyrics:
    song_id: int
    song_title: str
    lines: list[str]
    is_synced: bool

    def write_to_file(self, path: str):
        with open(path, 'w', encoding='utf-8', opener=opener) as f:
            f.writelines(self.lines)

    @classmethod
    def from_dict(cls, song_id: int, song_title: str, lyrics_dict: dict[int, str]) -> 'Lyrics':
        """
        Constructs a synced lyrics object from a dict of timestamps/lines.
        """
        return cls(song_id=song_id, song_title=song_title, lines=_convert_to_lrc(lyrics_dict), is_synced=True)


def _convert_to_lrc(lyrics_dict: dict[int, str]) -> list[str]:
    """
    Takes a dict of millis/lines and converts it to the lrc format.
    """
    sorted_lines = sorted(lyrics_dict.items())
    formatted_lines = [f'[{format_lrc_timestamp(timestamp)}]{line}\n' for timestamp, line in sorted_lines]
    return formatted_lines



import os
import re
from os import path
from pathlib import Path
from sys import stderr


def fix_synced_lyrics(collection_path: Path):
    # Validate collection path
    if not collection_path.is_dir():
        print(f'Error: directory \'{collection_path}\' does not exist', file=stderr)
        exit(2)

    timestamp_pattern = re.compile(r'^\[(\d+):(\d{2}):(\d{2})]', flags=re.MULTILINE)

    for root_dir, dirs, files in os.walk(collection_path, topdown=True):
        if path.exists(path.join(root_dir, '.nolyrics')):
            dirs.clear()
            continue

        for file in files:
            extension = path.splitext(file)[1].lower()
            if extension == '.lrc':
                fix_synced_lyrics_file(root_dir, file, timestamp_pattern)


def fix_synced_lyrics_file(dirname: str, filename: str, timestamp_pattern: re.Pattern):
    filepath = path.join(dirname, filename)

    with open(filepath, 'r', encoding='utf-8') as f:
        file_content = f.read()

    new_content = timestamp_pattern.sub(r'[\1:\2.\3]', file_content)
    if new_content == file_content:
        return

    print(f'Fixing synced lyrics format for "{filename}"')
    with open(filepath, 'w', encoding='utf-8') as f:
        f.write(new_content)


def format_lrc_timestamp(millis: int) -> str:
    """
    Format a given milliseconds value into the LRC timestamp format (MM:SS.cc).
    """
    minutes = millis // 60000
    seconds = (millis % 60000) // 1000
    centis = (millis % 1000) // 10
    return f'{minutes:02d}:{seconds:02d}.{centis:02d}'



provider_registry = {}


def register_provider(*names: str):
    """
    Decorator to register a provider class with one or more names.
    """

    def wrapper(cls):
        for name in names:
            provider_registry[name] = cls
        return cls

    return wrapper



import json
from dataclasses import dataclass
from json.decoder import JSONDecodeError
from urllib.parse import unquote

import httpx
from httpx import RequestError

from lyriks.lyrics import Lyrics
from lyriks.providers import Song

GENIE_ALBUM_API_URL = 'https://app.genie.co.kr/song/j_AlbumSongList.json?axnm={album_id:d}'
GENIE_LYRICS_API_URL = 'https://dn.genie.co.kr/app/purchase/get_msl.asp?songid={song_id:d}&callback=GenieCallback'
GENIE_STREAM_INFO_API_URL = 'https://stm.genie.co.kr/player/j_StmInfo.json?xgnm={song_id:d}'

CURL_USER_AGENT = 'curl/8.7.1'  # for whatever reason, this works, but the python-requests UA doesn't


@dataclass
class GenieSong(Song):
    pass


def get_album_songs(album_id: int) -> list[GenieSong] | None:
    try:
        response = httpx.get(
            GENIE_ALBUM_API_URL.format(album_id=album_id),
            headers={'User-Agent': CURL_USER_AGENT},
        ).json()
    except RequestError | JSONDecodeError:
        return None

    try:
        songs = list(response['DATA1']['DATA'])
    except KeyError:
        return None

    # Reject albums with multiple CDs
    if any(song.get('ALBUM_CD_NO') != '1' for song in songs):
        return None

    # Extract songs
    result = []

    for song in songs:
        song_id = song.get('SONG_ID')
        track_num = song.get('ALBUM_TRACK_NO')
        song_name = song.get('SONG_NAME')

        if song_id is None or track_num is None or song_name is None:
            return None

        try:
            song_id = int(song_id)
            track_num = int(track_num)
        except ValueError:
            return None

        result.append(GenieSong(id=song_id, album_index=track_num, title=unquote(song_name)))

    result = sorted(result, key=lambda x: x.album_index)

    return result


def get_stream_info(song_id: int) -> dict | None:
    try:
        response = httpx.get(
            GENIE_STREAM_INFO_API_URL.format(song_id=song_id),
            headers={'User-Agent': CURL_USER_AGENT},
        ).json()
    except RequestError | JSONDecodeError:
        return None

    try:
        stream_info = response['DataSet']['DATA'][0]
    except KeyError:
        return None

    return stream_info


def get_song_lyrics(song: GenieSong) -> Lyrics | None:
    # Try to fetch synced lyrics
    try:
        response = httpx.get(
            GENIE_LYRICS_API_URL.format(song_id=song.id),
            headers={'User-Agent': CURL_USER_AGENT},
        ).text
    except RequestError:
        return None

    if response is not None and response.startswith('GenieCallback('):
        # We (probably) got synced lyrics
        response = response.removeprefix('GenieCallback(').removesuffix(');')
        try:
            raw_lyrics = json.loads(response)
        except JSONDecodeError:
            return None

        # Convert timestamps and cleanup lines
        lyrics_dict: dict[int, str] = {int(timestamp): line.strip() for timestamp, line in raw_lyrics.items()}

        return Lyrics.from_dict(song.id, song.title, lyrics_dict)
    else:
        # Fall back to static lyrics from stream info
        stream_info = get_stream_info(song.id)
        if not stream_info:
            return None

        try:
            raw_lyrics = stream_info['LYRICS']
        except KeyError:
            return None

        # Reject empty lyrics
        if not raw_lyrics:
            return None

        lines = unquote(raw_lyrics).split('<br>')

        # Reject instrumental tracks
        if '이 곡은 연주곡 입니다.' in lines:
            return None

        return Lyrics(song_id=song.id, song_title=song.title, lines=lines, is_synced=False)



from lyriks import Provider
from lyriks.lyrics import Lyrics
from lyriks.mb_client import Release
from .api import vibe_api
from .api.vibe_api import VibeSong
from .registry import register_provider


@register_provider('vibe')
class Vibe(Provider):
    """
    Provider for Naver Vibe.
    """

    def fetch_recording_lyrics(self, track_release: Release, recording_mbid: str) -> Lyrics | None:
        # Resolve album
        vibe_songs = self.get_mapped_provider_songs(
            track_release,
            lambda r: r.extract_url_id(r'https://vibe.naver.com/album/(\d+)'),
            vibe_api.get_album_songs,
        )
        if not vibe_songs:
            return None

        # Get song for recording
        vibe_song = vibe_songs.get(recording_mbid)
        if not vibe_song:
            return None

        # Fetch lyrics
        return self.fetch_provider_song_lyrics(vibe_song)

    def fetch_song_by_id(self, song_id: int) -> VibeSong | None:
        return vibe_api.get_song_info(song_id)

    def fetch_provider_song_lyrics(self, song: VibeSong) -> Lyrics | None:
        return vibe_api.get_song_lyrics(song)

    @property
    def provider_domain(self) -> str:
        return 'vibe.naver.com'



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Callable, TypeVar

from lyriks.lyrics import Lyrics
from lyriks.mb_client import Release, Artist
from .util import pick_release_from_release_group


@dataclass
class Song(ABC):
    """
    Represents a song in a provider-specific context.
    """

    id: int
    album_index: int
    title: str


T = TypeVar('T')
S = TypeVar('S', bound=Song)


class Provider(ABC):
    """
    Abstract base class for lyrics providers.
    Supports fetching lyrics as part of the sync process,
    or once for a single song, identified by its provider-specific ID.

    The provider is responsible for caching releases to avoid redundant API calls.
    Artists and releases that don't have a URL relationship can be recorded for later reporting.
    """

    def __init__(self):
        self.cache: dict[str, dict[str, object] | None] = {}
        self.missing_artists: dict[str, Artist] = {}
        self.missing_releases: dict[str, Release] = {}

    @abstractmethod
    def fetch_recording_lyrics(self, track_release: Release, recording_mbid: str) -> Lyrics | None:
        """
        Fetch lyrics for a track, identified by its recording MBID and the release it appears on.

        Typically, this resolves the provider-specific song entity and then
        delegates to fetch_provider_song_lyrics to fetch the lyrics.
        """
        pass

    @abstractmethod
    def fetch_song_by_id(self, song_id: int) -> Song | None:
        """
        Fetch a provider-specific song entity by its ID.
        """
        pass

    @abstractmethod
    def fetch_provider_song_lyrics(self, song: Song) -> Lyrics | None:
        """
        Fetch lyrics for a given song entity. Its content may be provider-specific.
        """
        pass

    @property
    @abstractmethod
    def provider_domain(self) -> str:
        """
        Get the primary domain of the provider.
        """
        pass

    def has_artist_url(self, artist: Artist) -> bool:
        """
        Check if the artist has a URL relationship for the service used by this provider.
        :return: True if we're unable to check or if this artist has a URL, False otherwise.
        """

        domain = self.provider_domain
        if any(domain in url for url in artist.urls):
            return True

        if artist.id not in self.missing_artists:
            print(f'\rNo {domain} URL found for artist {artist.name} [{artist.id}]')
            self.missing_artists[artist.id] = artist

        return False

    def get_mapped_provider_songs(
        self, track_release: Release, selector: Callable[[Release], T | None], fetcher: Callable[[T], list[S] | None]
    ) -> dict[str, S] | None:
        """
        Get songs for a track release, matched to its recordings.

        :param track_release: The release to fetch and map songs for.
        :param selector: A lambda function extracting a provider-specific identifier from a release if available.
        :param fetcher: A function that takes the provider-specific identifier and returns a list of songs.
        :return: A dictionary mapping recording MBIDs to provider-specific songs, or None if there was an error.
        """
        if track_release.id in self.cache:
            return self.cache[track_release.id]

        result = pick_release_from_release_group(track_release, selector)
        if not result:
            print(f'\rNo URL found for release {track_release.title} [{track_release.id}]')
            self.cache[track_release.id] = None
            self.missing_releases[track_release.id] = track_release
            return None
        matched_release, album_id = result

        provider_songs = fetcher(album_id)
        if not provider_songs:
            self.cache[track_release.id] = None
            return None

        # Ensure track count matches
        if len(provider_songs) != matched_release.get_track_count():
            print(f'\rTrack count mismatch for release {track_release.title} [{track_release.id}]')
            self.cache[track_release.id] = None
            return None

        # Match recordings to songs
        mapped_songs = {}

        # Iterate over all tracks in the release
        for medium in matched_release.media:
            for track in medium['tracks']:
                recording_mbid: str = track['recording']['id']
                try:
                    # Match song by track number if possible
                    track_number = int(track['number'])
                    song = next(song for song in provider_songs if song.album_index == track_number)
                except ValueError | StopIteration:
                    # Fall back to track position
                    track_index = track['position'] - 1
                    if track_index >= len(provider_songs):
                        continue
                    song = provider_songs[track_index]

                mapped_songs[recording_mbid] = song

        self.cache[track_release.id] = mapped_songs

        return mapped_songs



from pathlib import Path

import click

from .const import PROGNAME, VERSION
from .default_group import DefaultGroup
from .lyrics_fetcher import main, fetch_single_song
from .provider_choice import ProviderChoice
from .providers import Provider
from .providers.genie import Genie as Genie
from .providers.qqm import QQMusic as QQMusic
from .providers.vibe import Vibe as Vibe
from .util import fix_synced_lyrics


@click.group(
    cls=DefaultGroup,
    default_command='sync',
    no_args_is_help=True,
    invoke_without_command=True,
    context_settings=dict(
        ignore_unknown_options=True,
        max_content_width=160,
    ),
)
def cli():
    pass


@cli.command()
@click.option(
    '-a',
    '--check-artist',
    is_flag=True,
    help='ensure artist has a Genie URL when processing albums',
)
@click.option(
    '-n',
    '--dry-run',
    is_flag=True,
    help='fetch lyrics without writing them to files',
)
@click.option(
    '-u',
    '--upgrade',
    is_flag=True,
    help='upgrade existing static lyrics to synced lyrics if possible',
)
@click.option(
    '-f',
    '--force',
    is_flag=True,
    help=(
        'force fetching lyrics for all tracks, even if they already have them'
        ' - THIS WILL OVERWRITE EXISTING LYRICS FILES!'
    ),
)
@click.option(
    '-I',
    '--skip-instrumentals',
    is_flag=True,
    help='skip instrumental tracks',
)
@click.option(
    '-R',
    '--report',
    'report_path',
    is_flag=False,
    flag_value='report.html',
    type=click.Path(),
    help=(
        'write a HTML report of releases missing album URLs to a file at PATH'
        ' (default: report.html in the current directory)'
    ),
)
@click.option(
    '-P',
    '--provider',
    type=ProviderChoice(),
    default='genie',
    help='the lyrics provider to use (default: genie)',
)
@click.argument('collection_path', type=click.Path(exists=True))
@click.version_option(
    VERSION,
    '-v',
    '--version',
    prog_name=PROGNAME,
    message='%(prog)s %(version)s',
    help='show the version and exit',
)
@click.help_option(
    '-h',
    '--help',
    help='show this message and exit',
)
def sync(
    check_artist: bool,
    dry_run: bool,
    upgrade: bool,
    force: bool,
    skip_instrumentals: bool,
    report_path: str,
    provider: Provider,
    collection_path: str,
):
    """
    A command line tool that fetches lyrics from Genie.
    """
    report_path = Path(report_path) if report_path else None
    collection_path = Path(collection_path)
    main(provider, check_artist, dry_run, upgrade, force, skip_instrumentals, report_path, collection_path)


@cli.command()
@click.option(
    '-P',
    '--provider',
    type=ProviderChoice(),
    default='genie',
    help='the lyrics provider to use (default: genie)',
)
@click.option(
    '-o',
    '--output',
    'output_path',
    type=click.Path(),
    help='write the lyrics to PATH (default: <song title>.<ext> in the current directory)',
)
@click.argument('song_id', type=int)
@click.help_option(
    '-h',
    '--help',
    help='show this message and exit',
)
def fetch(provider: Provider, output_path: str, song_id: int):
    """
    Fetch lyrics for a single song from Genie.

    The song ID can be found in the URL of the song's Genie page.
    """
    fetch_single_song(provider, song_id, output_path)


@cli.command()
@click.argument('collection_path', type=click.Path(exists=True))
def fix(collection_path: str):
    """
    Fix the format of synced lyrics in the collection.

    Specifically, it replaces timestamps in the previously used format [mm:ss:xx] with [mm:ss.xx].
    """
    collection_path = Path(collection_path)
    fix_synced_lyrics(collection_path)



from lyriks.lyrics import Lyrics
from lyriks.mb_client import Release
from . import Provider
from .api import genie_api
from .api.genie_api import GenieSong
from .registry import register_provider


@register_provider('genie')
class Genie(Provider):
    """
    Provider for Genie Music.
    """

    def fetch_recording_lyrics(self, track_release: Release, recording_mbid: str) -> Lyrics | None:
        # Resolve Genie album
        genie_songs = self.get_mapped_provider_songs(
            track_release,
            lambda r: r.extract_url_id(r'https://(?:www.)?genie.co.kr/detail/albumInfo\?axnm=(\d+).*'),
            genie_api.get_album_songs,
        )
        if not genie_songs:
            return None

        # Get Genie song for recording
        genie_song = genie_songs.get(recording_mbid)
        if not genie_song:
            return None

        # Fetch lyrics
        return self.fetch_provider_song_lyrics(genie_song)

    def fetch_song_by_id(self, song_id: int) -> GenieSong | None:
        stream_info = genie_api.get_stream_info(song_id)
        if stream_info is None:
            return None

        try:
            album_id = int(stream_info['ALBUM_ID'])
        except KeyError | ValueError:
            return None

        genie_songs = genie_api.get_album_songs(album_id)
        if not genie_songs:
            return None

        return next((s for s in genie_songs if s.id == song_id), None)

    def fetch_provider_song_lyrics(self, song: GenieSong) -> Lyrics | None:
        return genie_api.get_song_lyrics(song)

    @property
    def provider_domain(self) -> str:
        return 'genie.co.kr'



import json
import random
import re
import zlib
from dataclasses import dataclass, field
from datetime import datetime
from json import JSONDecodeError

import httpx
import lxml.etree as xml
import pyqqmusicdes
from httpx import RequestError
from lxml.etree import XMLParser

from lyriks.lib.zzc_sign import zzc_sign
from lyriks.lyrics import Lyrics
from lyriks.providers import Song
from lyriks.util import format_lrc_timestamp

xml.set_default_parser(XMLParser(no_network=True, recover=True, remove_blank_text=True))

QQM_API_URL = "https://u.y.qq.com/cgi-bin/musicu.fcg"
QQM_COMM = {
    'cv': 4747474,
    'ct': 24,
    'format': 'json',
    'inCharset': 'utf-8',
    'outCharset': 'utf-8',
    'notice': 0,
    'platform': 'yqq.json',
    'needNewCode': 1,
    'uin': ''.join(random.sample('1234567890', 10)),
    'g_tk_new_20200303': 1077614320,
    'g_tk': 1077614320,
}
QQM_LYRICS_API_URL = "https://c.y.qq.com/qqmusic/fcgi-bin/lyric_download.fcg"
QQM_DES_KEY = b'!@#)(*$%123ZXC!@!@#)(NHL'

CHROME_USER_AGENT = (
    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/130.0.0.0 Safari/537.36"
)

LYRIC_CONTENT_TAG = 'LyricContent'
LYRIC_CONTENT_REGEX = re.compile(r'LyricContent="([\s\S]*?)"\s*/>')


@dataclass
class QQMSong(Song):
    mid: str
    artists: list[str] = field(default_factory=list)

    @classmethod
    def from_song_info(cls, song_info: dict) -> 'QQMSong':
        try:
            song_id = song_info['id']
            song_mid = song_info['mid']
            album_index = song_info['index_album']
            title = song_info['title']
            artists = [artist['name'] for artist in song_info['singer']]
        except KeyError:
            raise ValueError("Invalid song info data")

        return cls(id=song_id, mid=song_mid, album_index=album_index, title=title, artists=artists)


def _qqm_request(modules: list[dict]) -> list[dict]:
    request = dict([('comm', QQM_COMM)] + [(f'req_{i + 1}', module) for i, module in enumerate(modules)])
    body = json.dumps(request)
    signature = zzc_sign(body)

    try:
        response = httpx.post(
            QQM_API_URL,
            params={
                "_": int(datetime.now().timestamp() * 1000),
                "sign": signature,
            },
            headers={
                "Accept": "application/json",
                "Accept-Language": "en-DE,en;q=1",
                "Content-Type": "application/x-www-form-urlencoded",
                "Origin": "https://y.qq.com",
                "Referer": "https://y.qq.com/",
                "User-Agent": CHROME_USER_AGENT,
            },
            content=body,
        ).json()
    except RequestError | JSONDecodeError:
        return []

    try:
        response_modules = [response[f'req_{i + 1}'] for i in range(len(modules))]
    except KeyError:
        return []

    return response_modules


def get_album_songs(album_mid: str) -> list[QQMSong]:
    response = _qqm_request(
        [
            {
                'module': 'music.musichallAlbum.AlbumSongList',
                'method': 'GetAlbumSongList',
                'param': {'albumMid': album_mid, 'albumID': 0, 'begin': 0, 'num': 100, 'order': 2},
            }
        ]
    )

    if not response:
        return []

    try:
        song_infos = [song['songInfo'] for song in response[0]['data']['songList']]
    except IndexError | KeyError:
        return []

    try:
        return [QQMSong.from_song_info(song_info) for song_info in song_infos]
    except ValueError:
        return []


def get_song_info(song_id: int) -> QQMSong | None:
    response = _qqm_request(
        [
            {
                "module": "music.trackInfo.UniformRuleCtrl",
                "method": "CgiGetTrackInfo",
                "param": {"ids": [song_id], "types": [0]},
            }
        ]
    )

    if not response:
        return None

    try:
        song_info = response[0]['data']['tracks'][0]
    except IndexError | KeyError:
        return None

    return QQMSong.from_song_info(song_info)


def get_song_lyrics(song: QQMSong) -> Lyrics | None:
    request = {
        "version": "15",
        "miniversion": "82",
        "lrctype": "4",
        "musicid": song.id,
    }

    try:
        response = httpx.post(
            QQM_LYRICS_API_URL,
            headers={
                "Accept": "application/xml",
                "Content-Type": "application/x-www-form-urlencoded",
                "Referer": "https://c.y.qq.com/",
                "User-Agent": CHROME_USER_AGENT,
            },
            data=request,
        ).text
    except RequestError:
        return None

    if response is None:
        return None

    # Remove surrounding comment tags
    response_text = response.replace("<!--", "").replace("-->", "")

    # Parse main XML response
    root = xml.fromstring(response_text)
    lyric_node = root.find(".//lyric")
    if lyric_node is None:
        return None

    content_node = lyric_node.find('content')
    if content_node is None:
        return None

    content_text = _decrypt_content_node(content_node)

    # The XML is malformed (unescaped quotes inside attribute) and contains newlines.
    # Thus, as standard XML parsers would fail, we use a regex to extract the content.
    match = LYRIC_CONTENT_REGEX.search(content_text)
    if not match:
        return None

    lyric_content = match.group(1)
    lines = lyric_content.splitlines()
    lines = _convert_qrc_to_lrc(lines)

    return Lyrics(song_id=song.id, song_title=song.title, lines=lines, is_synced=True)


def _decrypt_content_node(content_node) -> str | None:
    content = content_node.text
    if content is None:
        return None

    content = content.strip()
    if len(content) == 0:
        return None

    buf = bytes.fromhex(content)

    res = pyqqmusicdes.decrypt_des(buf, QQM_DES_KEY)
    if res != 0:
        return None

    buf = zlib.decompress(buf)

    try:
        content_text = buf.decode('utf-8')
    except UnicodeDecodeError:
        return None

    return content_text


def _convert_qrc_to_lrc(lines: list[str]) -> list[str] | None:
    """
    Converts lines from QRC format to LRC format.
    """
    lrc_lines = []
    metadata_regex = re.compile(r'\[[a-z]+:[^]]*]')
    line_timestamp_regex = re.compile(r'\[(\d+),(\d+)]')
    words_regex = re.compile(r'(.*?)\((\d+),(\d+)\)')

    for line in lines:
        if not line.startswith('['):
            # Ignore lines without a start tag
            continue

        if metadata_regex.match(line):
            # Copy metadata lines as-is
            lrc_lines.append(f'{line}\n')
            continue

        line_timestamp_match = line_timestamp_regex.match(line)
        if not line_timestamp_match:
            # Abort on malformed lines
            return None

        line_start = int(line_timestamp_match.group(1))
        lrc_line_timestamp = f'[{format_lrc_timestamp(line_start)}]'

        line_content_start = line_timestamp_match.end()
        line_content = line[line_content_start:]

        timed_words = words_regex.findall(line_content)
        lrc_line = ''
        for word, start_str, duration_str in timed_words:
            start = int(start_str)
            end = start + int(duration_str)
            lrc_line += f'<{format_lrc_timestamp(start)}>{word}<{format_lrc_timestamp(end)}>'

        lrc_lines.append(f'{lrc_line_timestamp}{lrc_line}\n')

    return lrc_lines



import html
import os
from os import PathLike
from os import path
from pathlib import Path
from sys import stderr

import mutagen
from mutagen.easymp4 import EasyMP4Tags

from .mb_client import Artist, Release, get_artist, get_release_by_track
from .providers import Provider

TITLE_TAG = 'title'
ALBUM_TAG = 'album'
TRACKNUMBER_TAG = 'tracknumber'
ALBUMARTIST_TAG = 'albumartist'
MB_RGID_TAG = 'musicbrainz_releasegroupid'
MB_RTID_TAG = 'musicbrainz_releasetrackid'
MB_AAID_TAG = 'musicbrainz_albumartistid'

VARIOUS_ARTISTS_MBID = '89ad4ac3-39f7-470e-963a-56509c546377'

EasyMP4Tags.RegisterFreeformKey(MB_RGID_TAG, 'MusicBrainz Release Group Id')
EasyMP4Tags.RegisterFreeformKey(MB_RTID_TAG, 'MusicBrainz Release Track Id')


def main(
    provider: Provider,
    check_artist: bool,
    dry_run: bool,
    upgrade: bool,
    force: bool,
    skip_instrumentals: bool,
    report_path: Path | None,
    collection_path: Path,
):
    # Validate collection path
    if not collection_path.is_dir():
        print(f'Error: directory \'{collection_path}\' does not exist', file=stderr)
        exit(2)

    # Normalize and validate report path
    if report_path:
        report_path = report_path.expanduser().absolute()
        if report_path.is_dir():
            report_path = report_path / 'report.html'
        if not report_path.parent.exists():
            print(f'Error: directory \'{report_path.parent}\' does not exist', file=stderr)
            exit(2)

    fetcher = LyricsFetcher(provider, check_artist, dry_run, upgrade, force, skip_instrumentals)

    for root_dir, dirs, files in os.walk(collection_path, topdown=True):
        if path.exists(path.join(root_dir, '.nolyrics')):
            dirs.clear()
            continue

        for file in files:
            extension = path.splitext(file)[1].lower()
            if extension in ('.flac', '.m4a', '.mp3'):
                fetcher.fetch_lyrics(root_dir, file)

    if report_path:
        try:
            fetcher.write_report(report_path)
        except OSError:
            print(f'Error: could not write report to \'{report_path}\'', file=stderr)
            exit(2)


def fetch_single_song(provider: Provider, song_id: int, output_path: str):
    song = provider.fetch_song_by_id(song_id)
    if song is None:
        print('Song not found.')
        return
    lyrics = provider.fetch_provider_song_lyrics(song)
    if lyrics is None:
        print('Failed to fetch lyrics.')
        return

    extension = 'lrc' if lyrics.is_synced else 'txt'
    output_path = output_path or f'{lyrics.song_title}.{extension}'
    lyrics.write_to_file(output_path)

    print(f'Lyrics saved to {output_path}')


class LyricsFetcher:
    def __init__(
        self,
        provider: Provider,
        check_artist: bool = False,
        dry_run: bool = False,
        upgrade: bool = False,
        force: bool = False,
        skip_inst: bool = False,
    ):
        self.provider = provider
        self.check_artist = check_artist
        self.dry_run = dry_run
        self.upgrade = upgrade
        self.force = force
        self.skip_inst = skip_inst
        self.artist_cache: dict[str, Artist] = {}
        self.release_cache: dict[str, Release] = {}

    def fetch_lyrics(self, dirname: str, filename: str) -> None:
        filepath = path.join(dirname, filename)
        basename = filename.rsplit('.', 1)[0]

        # Skip instrumental tracks if enabled and applicable
        if self.skip_inst and ('instrumental' in basename.lower() or 'inst.' in basename.lower()):
            return

        # Skip if .nolyrics file exists
        nolyrics_file = path.join(dirname, f'{basename}.nolyrics')
        if path.exists(nolyrics_file):
            return

        synced_lyrics_file = path.join(dirname, f'{basename}.lrc')
        static_lyrics_file = path.join(dirname, f'{basename}.txt')
        has_synced_lyrics = path.exists(synced_lyrics_file)
        has_static_lyrics = path.exists(static_lyrics_file)

        # Skip if lyrics already exist
        if (has_synced_lyrics or (has_static_lyrics and not self.upgrade)) and not self.force:
            return

        file = mutagen.File(filepath, easy=True)
        if not file:
            return

        tags = file.tags
        if (
            TITLE_TAG not in tags
            or ALBUM_TAG not in tags
            or TRACKNUMBER_TAG not in tags
            or MB_RGID_TAG not in tags
            or MB_RTID_TAG not in tags
        ):
            return

        title = tags[TITLE_TAG][0] or 'Unknown title'
        album = tags[ALBUM_TAG][0] or 'Unknown album'
        rg_mbid = tags[MB_RGID_TAG][0]
        track_mbid = tags[MB_RTID_TAG][0]

        # Handle empty MBIDs
        if not rg_mbid or not track_mbid:
            return

        # Check artist URL
        if self.check_artist and not self.has_artist_url(self.provider, tags):
            return

        # Resolve release for the track
        track_release = self.get_release(track_mbid, album)
        if not track_release:
            return

        # Resolve track
        for medium in track_release.get_track_map():
            track = medium.get(track_mbid)
            if track:
                break
        else:
            return

        # Fetch lyrics
        print(f'Fetching lyrics for {title}', end='', flush=True)
        recording_mbid = track['recording']['id']
        lyrics = self.provider.fetch_recording_lyrics(track_release, recording_mbid)
        if not lyrics:
            print(f'\rNo lyrics found for {title}')
            return

        if self.dry_run:
            print(f'\rFetching lyrics for {title} - done [dry run]')
        else:
            # Write lyrics to file
            if lyrics.is_synced:
                print(f'\rFetching lyrics for {title} - writing to {synced_lyrics_file}')
                lyrics.write_to_file(synced_lyrics_file)

                # Remove static lyrics file if necessary
                if has_static_lyrics:
                    os.unlink(static_lyrics_file)
            elif has_synced_lyrics:
                print(f'\rFetching lyrics for {title} - not writing static lyrics, synced lyrics already exist')
            elif self.upgrade and has_static_lyrics:
                print(f'\rFetching lyrics for {title} - no synced lyrics available to upgrade to')
            else:
                print(f'\rFetching lyrics for {title} - writing to {static_lyrics_file}')
                lyrics.write_to_file(static_lyrics_file)

    def has_artist_url(self, provider: Provider, tags) -> bool:
        """
        Check if the artist has a Genie URL.
        :return: True if we're unable to check or if this artist has a Genie URL, False otherwise.
        """
        if MB_AAID_TAG not in tags or ALBUMARTIST_TAG not in tags:
            return True

        albumartist_mbid = tags[MB_AAID_TAG][0]
        if not albumartist_mbid:  # handle empty MBID
            return True

        # Skip various artists
        if albumartist_mbid == VARIOUS_ARTISTS_MBID:
            return True

        albumartist = tags[ALBUMARTIST_TAG][0] or 'Unknown artist'
        artist = self.get_artist(albumartist_mbid, albumartist)
        if not artist or provider.has_artist_url(artist):
            return True

        return False

    def get_artist(self, artist_mbid: str, artist_name: str) -> Artist | None:
        if artist_mbid in self.artist_cache:
            return self.artist_cache[artist_mbid]

        print(f'Fetching artist info for {artist_name}', end='')

        artist = get_artist(artist_mbid)
        if not artist:
            print(' - no artist found')
            return None

        self.artist_cache[artist_mbid] = artist

        print()  # terminate line

        return artist

    def get_release(self, track_mbid: str, album_name: str) -> Release | None:
        if track_mbid in self.release_cache:
            return self.release_cache[track_mbid]

        print(f'Fetching release info for {album_name}', end='')

        release = get_release_by_track(track_mbid)
        if not release:
            print(' - no release found')
            return None

        for media in release.data['media']:
            for track in media['tracks']:
                self.release_cache[track['id']] = release

        print()  # terminate line

        return release

    # noinspection DuplicatedCode
    def write_report(self, file: str | PathLike[str]):
        with open(file, 'w') as f:
            f.write('<!DOCTYPE html>\n')
            f.write('<html>\n')
            f.write('<head><title>lyriks report</title></head>\n')
            f.write('<body>\n')
            f.write('<h1>lyriks report</h1>\n')
            f.write(f'<h2>Artists missing URLs ({len(self.provider.missing_artists)})</h2>\n')
            if self.provider.missing_artists:
                f.write('<ul>\n')
                for artist in self.provider.missing_artists.values():
                    url = f'https://musicbrainz.org/artist/{artist.id}'
                    f.write(f'<li><a href="{url}">{html.escape(artist.name)}</a></li>\n')
                f.write('</ul>\n')
            else:
                f.write('<p>None.</p>\n')
            f.write(f'<h2>Releases missing URLs ({len(self.provider.missing_releases)})</h2>\n')
            if self.provider.missing_releases:
                f.write('<ul>\n')
                for release in self.provider.missing_releases.values():
                    url = f'https://musicbrainz.org/release/{release.id}'
                    f.write(f'<li><a href="{url}">{html.escape(release.title)}</a></li>\n')
                f.write('</ul>\n')
            else:
                f.write('<p>None.</p>\n')
            f.write('</body>\n')
            f.write('</html>')



from dataclasses import dataclass, field
from json import JSONDecodeError

import httpx
from httpx import RequestError

from lyriks.lyrics import Lyrics
from lyriks.providers import Song

VIBE_LYRICS_API_URL = 'https://apis.naver.com/vibeWeb/musicapiweb/vibe/v4/lyric/{song_id:d}'
VIBE_ALBUM_TRACKS_API_URL = 'https://apis.naver.com/vibeWeb/musicapiweb/album/{album_id:d}/tracks?start=1&display=1000'
VIBE_TRACK_API_URL = 'https://apis.naver.com/vibeWeb/musicapiweb/track/{song_id:d}'


@dataclass
class VibeSong(Song):
    artists: list[str] = field(default_factory=list)

    @classmethod
    def from_track_info(cls, track_info: dict) -> 'VibeSong':
        try:
            track_id = track_info['trackId']
            album_index = track_info['trackNumber']
            title = track_info['trackTitle']
            artists = [artist['artistName'] for artist in track_info['artists']]
        except KeyError:
            raise ValueError('Invalid track info data')

        return cls(id=track_id, album_index=album_index, title=title, artists=artists)


def get_album_songs(album_id: int) -> list[VibeSong]:
    try:
        response = httpx.get(
            VIBE_ALBUM_TRACKS_API_URL.format(album_id=album_id),
            headers={'Accept': 'application/json'},
        ).json()
    except RequestError | JSONDecodeError:
        return []

    try:
        tracks = response['response']['result']['tracks']
    except KeyError:
        return []

    try:
        return [VibeSong.from_track_info(track_info) for track_info in tracks]
    except ValueError:
        return []


def get_song_info(song_id: int) -> VibeSong | None:
    try:
        response = httpx.get(
            VIBE_TRACK_API_URL.format(song_id=song_id),
            headers={'Accept': 'application/json'},
        ).json()
    except RequestError | JSONDecodeError:
        return None

    try:
        track_info = response['response']['result']['track']
    except KeyError:
        return None

    try:
        return VibeSong.from_track_info(track_info)
    except ValueError:
        return None


def get_song_lyrics(song: VibeSong) -> Lyrics | None:
    try:
        lyrics_response = httpx.get(
            VIBE_LYRICS_API_URL.format(song_id=song.id),
            headers={'Accept': 'application/json'},
        ).json()
    except RequestError | JSONDecodeError:
        return None

    try:
        lyrics_data = lyrics_response['response']['result']['lyric']
    except KeyError:
        return None

    if lyrics_data.get('hasSyncLyric'):
        try:
            sync_lyrics = lyrics_data['syncLyric']
            start_times = sync_lyrics['startTimeIndex']
            lines = sync_lyrics['contents'][0]['text']
        except IndexError | KeyError:
            return None

        try:
            joined_lyrics = list(zip(start_times, lines, strict=True))
        except ValueError:
            return None

        # Convert timestamps and cleanup lines
        lyrics_dict: dict[int, str] = {int(timestamp * 1000): line.strip() for timestamp, line in joined_lyrics}

        return Lyrics.from_dict(song.id, song.title, lyrics_dict)
    elif lyrics_data.get('hasNormalLyric'):
        try:
            lines = lyrics_data['normalLyric']['text'].split('\n')
        except KeyError:
            return None

        return Lyrics(song_id=song.id, song_title=song.title, lines=lines, is_synced=False)

    return None



# Taken and adapted from https://github.com/jixunmoe/qmweb-sign
# This file is not part of the project and licensed under the MIT License instead.

import re
from base64 import b64encode
from hashlib import sha1

PART_1_INDEXES = [23, 14, 6, 36, 16, 40, 7, 19]
PART_2_INDEXES = [16, 1, 32, 12, 19, 27, 8, 5]
SCRAMBLE_VALUES = [89, 39, 179, 150, 218, 82, 58, 252, 177, 52, 186, 123, 120, 64, 242, 133, 143, 161, 121, 179]

# JavaScript quirks emulation
PART_1_INDEXES = filter(lambda x: x < 40, PART_1_INDEXES)


def zzc_sign(payload: str) -> str:
    hash = sha1(payload.encode('utf-8')).hexdigest().upper()

    part1 = ''.join(map(lambda i: hash[i], PART_1_INDEXES))
    part2 = ''.join(map(lambda i: hash[i], PART_2_INDEXES))

    part3 = bytearray(20)
    for i, v in enumerate(SCRAMBLE_VALUES):
        value = v ^ int(hash[i * 2 : i * 2 + 2], 16)
        part3[i] = value
    b64_part = re.sub(rb'[\\/+=]', b'', b64encode(part3)).decode('utf-8')
    return f'zzc{part1}{b64_part}{part2}'.lower()



from click import Choice

from .providers.registry import provider_registry


class ProviderChoice(Choice):
    """
    A click Choice type to pick a provider from the registry.
    """

    def __init__(self):
        super().__init__(provider_registry.keys(), case_sensitive=False)

    def convert(self, value, param, ctx):
        # Validate and normalize allowed choice strings
        value = super().convert(value, param, ctx)

        provider_class = provider_registry.get(value)
        if provider_class is None:
            self.fail(
                self.get_invalid_choice_message(value=value, ctx=ctx),
                param=param,
                ctx=ctx,
            )

        return provider_class()



import typing as t

from click import Command, Context, Group
from click.utils import make_str


class DefaultGroup(Group):
    """
    A subclass of click.Group that provides a default command for the group.
    """

    def __init__(self, *args, **kwargs):
        self.default_command = kwargs.pop('default_command', None)
        super().__init__(*args, **kwargs)

    def resolve_command(
        self, ctx: Context, args: t.List[str]
    ) -> t.Tuple[t.Optional[str], t.Optional[Command], t.List[str]]:
        cmd_name = make_str(args[0])

        cmd = self.get_command(ctx, cmd_name)
        if cmd is not None:
            return cmd_name, cmd, args[1:]

        return None, self.get_default_command(), args

    def get_default_command(self) -> Command:
        if self.default_command is None:
            raise AttributeError('No default command set')
        return self.commands.get(self.default_command)

    def get_help(self, ctx: Context) -> str:
        return self.get_default_command().get_help(ctx)



import re
import time
from json.decoder import JSONDecodeError
from typing import AnyStr

import httpx

from .const import VERSION

API_URL = 'https://musicbrainz.org/ws/2'
USER_AGENT = f'lyriks/{VERSION} ( max@maxr1998.de )'
_ARTIST_INC = 'url-rels'
_RELEASE_INC = 'artist-credits+release-groups+recordings+media+url-rels'

last_request_time = 0


def handle_rate_limit():
    global last_request_time
    time_since = time.time() - last_request_time
    if time_since < 1:
        time.sleep(1 - time_since)
    last_request_time = time.time()


class Artist:
    def __init__(self, data: dict):
        self.data = data
        self.id: str = data['id']
        self.name: str = data['name']
        self.urls: set = {
            relation['url']['resource']
            for relation in data.get('relations', [])
            if relation.get('target-type') == 'url'
        }


class Release:
    def __init__(self, data: dict):
        self.data = data
        self.id: str = data['id']
        self.title: str = data['title']
        self.artist_credit: dict = data['artist-credit']
        self.rg_mbid: str = data['release-group']['id']
        self.media: list = self.data['media']

    def get_track_count(self) -> int:
        return sum(medium['track-count'] for medium in self.media)

    def get_track_map(self) -> list[dict[str, dict]]:
        return [{track['id']: track for track in medium['tracks']} for medium in self.media]

    def extract_url_str(self, pattern: AnyStr) -> str | None:
        """
        Extracts an ID from the release's URL relations using the provided RegEx pattern.
        The pattern must contain exactly one capturing group for the ID.
        """
        pattern_obj = re.compile(pattern)
        if pattern_obj.groups != 1:
            raise ValueError("Pattern must contain exactly one capturing group: %r" % pattern)

        relations = self.data['relations']
        for relation in relations:
            if relation.get('target-type') != 'url' or relation.get('ended'):
                continue
            url = relation['url']['resource']
            match = pattern_obj.fullmatch(url)
            if match:
                return match.group(1)
        return None

    def extract_url_id(self, pattern: AnyStr) -> int | None:
        """
        Extracts an integer ID from the release's URL relations using the provided RegEx pattern.
        The pattern must contain exactly one capturing group for the ID.
        """
        id_str = self.extract_url_str(pattern)
        if id_str is None:
            return None

        try:
            return int(id_str)
        except ValueError:
            return None


def get_artist(artist_mbid: str) -> Artist | None:
    handle_rate_limit()

    artist_url = f'{API_URL}/artist/{artist_mbid}?inc={_ARTIST_INC}'
    response = httpx.get(artist_url, headers={'User-Agent': USER_AGENT, 'Accept': 'application/json'})
    try:
        response_json = response.json()
    except JSONDecodeError:
        print(f'Error: could not fetch artist data for {artist_mbid}')
        return None

    if 'error' in response_json:
        print(f'Error: {response_json["error"]}')
        return None

    return Artist(response_json)


def get_releases(browse_url: str) -> list[Release]:
    handle_rate_limit()

    response = httpx.get(browse_url, headers={'User-Agent': USER_AGENT, 'Accept': 'application/json'})
    try:
        response_json = response.json()
    except JSONDecodeError:
        return []

    return [Release(release) for release in response_json.get("releases", [])]


def get_release_by_track(track_mbid: str) -> Release | None:
    return next(iter(get_releases(f'{API_URL}/release?track={track_mbid}&status=official&inc={_RELEASE_INC}')), None)


def get_releases_by_release_group(rg_mbid: str) -> list[Release]:
    return get_releases(f'{API_URL}/release?release-group={rg_mbid}&status=official&inc={_RELEASE_INC}')



from typing import Callable, TypeVar

from lyriks.mb_client import Release, get_releases_by_release_group

T = TypeVar('T')


def pick_release_from_release_group(
    release: Release, selector: Callable[[Release], T | None]
) -> tuple[Release, T] | None:
    """
    Pick a release from the release's release group that matches the given selector.

    Attempts to match the release itself first, then checks all releases in the release group,
    sorted by an increasing track count delta from the original release.
    This ensures the most similar release is chosen.
    """

    # Try to match the release itself first
    selection = selector(release)
    if selection is not None:
        return release, selection

    # If that fails, check all releases from the release group
    rg_releases = get_releases_by_release_group(release.rg_mbid)
    if not rg_releases:
        return None

    # Sort releases by track count delta
    track_release_track_count = release.get_track_count()
    rg_releases = sorted(rg_releases, key=lambda r: abs(r.get_track_count() - track_release_track_count))

    # Return the first matching release
    for rg_release in rg_releases:
        selection = selector(rg_release)
        if selection is not None:
            return rg_release, selection

    return None
